package com.metawear.mobility;

import android.Manifest;
import android.animation.ObjectAnimator;
import android.app.AlertDialog;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothManager;
import android.content.*;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.os.IBinder;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.WindowManager;
import android.view.animation.DecelerateInterpolator;
import android.widget.ImageView;
import android.widget.ProgressBar;
import android.widget.TextView;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.appcompat.widget.Toolbar;
import androidx.recyclerview.widget.DividerItemDecoration;
import androidx.recyclerview.widget.LinearLayoutManager;
import androidx.recyclerview.widget.RecyclerView;

import com.google.android.material.floatingactionbutton.FloatingActionButton;
import com.mbientlab.metawear.android.BtleService;

import java.util.Objects;

/**
 * Used permission check everywhere to make sure user won't accidentally disabled some of the permissions.
 */
public class MainActivity extends androidx.appcompat.app.AppCompatActivity implements ServiceConnection {
    private BtleService.LocalBinder serviceBinder;
    private MetawearServices metawearServices;

    private BluetoothAdapter bluetoothAdapter;
    final int REQUEST_ENABLE_BT = 1;
    final int PERMISSION_REQUEST_COARSE_LOCATION = 2;
    final int PERMISSION_REQUEST_WRITE_STORAGE = 3;
    final int PERMISSION_REQUEST_READ_STORAGE = 4;
    boolean bluetoothIsEnabled = false, locationIsEnabled = false;
    boolean readPermissionIsGranted = false, writePermissionIsGranted = false;

    private RecyclerView recyclerView;
    private RecyclerView.LayoutManager layoutManager;
    FloatingActionButton refreshButton;
    TextView loading;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        setupLayout();//Bind all xml layout and component, add decorations
        initialize(); //Check permissions and bind services

        // specify recycler listview adapter
        DeviceList deviceList = DeviceList.getInstance();
        recyclerView.setAdapter(deviceList.getListAdapter());

    }

    private void setupLayout() {
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON); //Keep screen awake
        recyclerView = findViewById(R.id.device_list);
        layoutManager = new LinearLayoutManager(this);
        recyclerView.setLayoutManager(layoutManager);
        recyclerView.addItemDecoration(new DividerItemDecoration(recyclerView.getContext(),
                LinearLayoutManager.VERTICAL));
        Toolbar toolbar = findViewById(R.id.toolbar);
        setSupportActionBar(toolbar);
        refreshButton = findViewById(R.id.refresh_button);
        refreshButton.setOnClickListener(v -> {
            BlueToothScanner blueToothScanner = new BlueToothScanner(this, bluetoothAdapter);
            refreshButton.hide();
            loading.setVisibility(View.VISIBLE);
            blueToothScanner.scanLeDevices();
        });
        refreshButton.setImageResource(R.drawable.ic_refresh_black_24dp);
        refreshButton.hide();
        loading = findViewById(R.id.loading_text);
        loading.setVisibility(View.VISIBLE);
    }

    private void initialize() {

        //Setup bluetooth, location and external storage access
        final BluetoothManager bluetoothManager =
                (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        bluetoothAdapter = bluetoothManager.getAdapter();

        bluetoothIsEnabled = bluetoothAdapter != null && bluetoothAdapter.isEnabled();
        locationIsEnabled = this.checkSelfPermission(Manifest.permission.ACCESS_COARSE_LOCATION) == PackageManager.PERMISSION_GRANTED;

        if (!(bluetoothIsEnabled && locationIsEnabled)) {
            final AlertDialog.Builder builder = new AlertDialog.Builder(this);
            builder.setMessage("Please enable bluetooth and grant location access to connect to MetaWear." +
                    " If you want to download data to your mobile device, please grant external storage access");
            builder.setPositiveButton(android.R.string.ok, null);
            builder.setOnDismissListener(dialogInterface -> {
                if (!bluetoothIsEnabled) {
                    Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
                    startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT);
                }
                if (!locationIsEnabled) {
                    requestPermissions(new String[]{Manifest.permission.ACCESS_COARSE_LOCATION}, PERMISSION_REQUEST_COARSE_LOCATION);
                }

            }).show();
        } else {
            BlueToothScanner blueToothScanner = new BlueToothScanner(this, bluetoothAdapter);
            blueToothScanner.scanLeDevices();
        }


        // Bind the service when the activity is created
        getApplicationContext().bindService(new Intent(this, BtleService.class),
                this, Context.BIND_AUTO_CREATE);
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Inflate the menu; this adds items to the action bar if it is present.
        getMenuInflater().inflate(R.menu.menu_toolbar, menu);
        return true;
    }

    /**
     * apply different actions depending on the menu item selected
     *
     * @param item The selected menu item
     * @return return true whe click event is finished
     */
    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        // Handle action bar item clicks.
        int id = item.getItemId();

        if (id == R.id.action_sync_all) {
            //Start syncing
            Thread thread = new Thread(this::downloadDataFromDevices);
            thread.start();
            return true;
        }

        if (id == R.id.action_start_all_metawear) {
            startDevices();
            return true;
        }

        if (id == R.id.action_stop_all_metawear) {
            stopDevices();
            return true;
        }
        return true;
    }

    /**
     * Bluetooth enable result
     */
    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data) {
        super.onActivityResult(requestCode, resultCode, data);
        if (requestCode == REQUEST_ENABLE_BT) {
            bluetoothIsEnabled = true;
            if (locationIsEnabled) {
                BlueToothScanner blueToothScanner = new BlueToothScanner(this, bluetoothAdapter);
                refreshButton.hide();
                loading.setVisibility(View.VISIBLE);
                blueToothScanner.scanLeDevices();
            }
        }
    }

    /**
     * Location access grant result
     */
    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String[] permissions, @NonNull int[] grantResults) {
        //Asks the user to grant the proper permissions required for BLE scanning (Location)
        if (requestCode == PERMISSION_REQUEST_COARSE_LOCATION) {
            if (grantResults[0] != PackageManager.PERMISSION_GRANTED) {
                Toast.makeText(this, "Location information is needed to search for Meatawear!", Toast.LENGTH_LONG).show();
            } else {
                locationIsEnabled = true;
                if (bluetoothIsEnabled) {
                    BlueToothScanner blueToothScanner = new BlueToothScanner(this, bluetoothAdapter);
                    refreshButton.hide();
                    loading.setVisibility(View.VISIBLE);
                    blueToothScanner.scanLeDevices();
                }
            }
        }

        if (requestCode == PERMISSION_REQUEST_READ_STORAGE) {
            if (grantResults[0] != PackageManager.PERMISSION_GRANTED) {
                Toast.makeText(this, "You cannot download data if you don't grant external storage access!", Toast.LENGTH_LONG).show();
            } else {
                readPermissionIsGranted = true;
                if (writePermissionIsGranted) {
                    startDevices();
                } else {
                    requestPermissions(new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE}, PERMISSION_REQUEST_WRITE_STORAGE);
                }
            }
        }

        if (requestCode == PERMISSION_REQUEST_WRITE_STORAGE) {
            if (grantResults[0] != PackageManager.PERMISSION_GRANTED) {
                Toast.makeText(this, "You cannot download data if you don't grant external storage access!", Toast.LENGTH_LONG).show();
            } else {
                writePermissionIsGranted = true;
                if (readPermissionIsGranted) {
                    startDevices();
                }
            }
        }
    }

    @Override
    public void onDestroy() {
        super.onDestroy();

        // Unbind the service when the activity is destroyed
        getApplicationContext().unbindService(this);
    }

    /**
     * Bind services and initialize metawear services
     */
    @Override
    public void onServiceConnected(ComponentName name, IBinder service) {
        // Typecast the binder to the service's LocalBinder class
        serviceBinder = (BtleService.LocalBinder) service;
        metawearServices = MetawearServices.getInstance();
        metawearServices.setupMetawearServices(this, bluetoothAdapter, serviceBinder);
    }

    @Override
    public void onServiceDisconnected(ComponentName componentName) {
    }

    /**
     * update battery level of indicated list item
     *
     * @param index        the index of indicated list item
     * @param batteryLevel new battery level
     */
    void updateBatteryLevel(int index, int batteryLevel) {
        View v = Objects.requireNonNull(recyclerView.getLayoutManager()).findViewByPosition(index);
        TextView batteryPercentage = Objects.requireNonNull(v).findViewById(R.id.battery_level_percentage);
        batteryPercentage.setText(batteryLevel + "%");

        ProgressBar batteryCircle = v.findViewById(R.id.battery_level_progress);
        batteryCircle.setVisibility(View.VISIBLE);
        ImageView profileImg = v.findViewById(R.id.user_image);
        TextView initials = v.findViewById(R.id.initials);
        profileImg.setVisibility(View.INVISIBLE);
        initials.setVisibility(View.INVISIBLE);

        ObjectAnimator animation = ObjectAnimator.ofInt(
                batteryCircle, "progress", batteryLevel);
        animation.setDuration(2000); // in milliseconds
        animation.setInterpolator(new DecelerateInterpolator());
        animation.start();
    }

    /**
     * Since I don't want my device to run all the tasks simultaneously, I setup a flag to inform the system when the last task is finished.
     * The system will instantly run to check if last task is finished and will stop when all the metawears are processed
     */
    void startDevices() {
        readPermissionIsGranted = checkSelfPermission(Manifest.permission.READ_EXTERNAL_STORAGE) == PackageManager.PERMISSION_GRANTED;
        writePermissionIsGranted = checkSelfPermission(Manifest.permission.WRITE_EXTERNAL_STORAGE) == PackageManager.PERMISSION_GRANTED;
        if (!readPermissionIsGranted || !writePermissionIsGranted) {
            requestPermissions(new String[]{Manifest.permission.READ_EXTERNAL_STORAGE}, PERMISSION_REQUEST_READ_STORAGE);
        } else {
            int index = 0;
            while (true) {
                if (MetawearServices.lastTaskIsFinished) {
                    MetawearServices.lastTaskIsFinished = false;
                    MetawearServices.getInstance().retrieveBoard(index, Operation.START_DEVICE, null);
                    index++;
                }

                if (index == DeviceList.getInstance().getMetawearList().size())
                    break;

                try {
                    Thread.sleep(500);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    void stopDevices() {
        int index = 0;
        while (true) {
            if (MetawearServices.lastTaskIsFinished) {
                MetawearServices.lastTaskIsFinished = false;
                MetawearServices.getInstance().retrieveBoard(index, Operation.STOP_DEVICE, null);
                index++;
            }

            if (index == DeviceList.getInstance().getMetawearList().size())
                break;

            try {
                Thread.sleep(500);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * just like startDevice function: flag, break, while loop, all the things
     */
    void downloadDataFromDevices() {
        readPermissionIsGranted = checkSelfPermission(Manifest.permission.READ_EXTERNAL_STORAGE) == PackageManager.PERMISSION_GRANTED;
        writePermissionIsGranted = checkSelfPermission(Manifest.permission.WRITE_EXTERNAL_STORAGE) == PackageManager.PERMISSION_GRANTED;
        if (!readPermissionIsGranted || !writePermissionIsGranted) {
            requestPermissions(new String[]{Manifest.permission.READ_EXTERNAL_STORAGE}, PERMISSION_REQUEST_READ_STORAGE);
        } else {
            int index = 0;
            while (true) {
                if (MetawearServices.lastTaskIsFinished) {
                    View v = Objects.requireNonNull(recyclerView.getLayoutManager()).findViewByPosition(index);
                    ProgressBar progress = Objects.requireNonNull(v).findViewById(R.id.sync_progress);
                    MetawearServices.lastTaskIsFinished = false;
                    MetawearServices.getInstance().retrieveBoard(index, Operation.SYNC_DEVICE, progress);
                    index++;
                }

                if (index == DeviceList.getInstance().getMetawearList().size())
                    break;

                try {
                    Thread.sleep(500);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }

    }

}