package com.metawear.mobility;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.os.Build;
import android.os.Environment;
import android.util.Log;
import android.view.View;
import android.widget.ProgressBar;
import android.widget.Toast;

import com.mbientlab.metawear.Data;
import com.mbientlab.metawear.MetaWearBoard;
import com.mbientlab.metawear.Route;
import com.mbientlab.metawear.Subscriber;
import com.mbientlab.metawear.android.BtleService;
import com.mbientlab.metawear.data.Acceleration;
import com.mbientlab.metawear.data.AngularVelocity;
import com.mbientlab.metawear.module.AccelerometerBmi160;
import com.mbientlab.metawear.module.AccelerometerBosch;
import com.mbientlab.metawear.module.GyroBmi160;
import com.mbientlab.metawear.module.Led;
import com.mbientlab.metawear.module.Logging;
import com.mbientlab.metawear.module.Settings;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import bolts.Continuation;
import bolts.Task;

import static androidx.constraintlayout.widget.Constraints.TAG;

class MetawearServices {

    private static MetawearServices metawearServices;

    private BluetoothAdapter bluetoothAdapter;
    private MetaWearBoard board;
    private AccelerometerBmi160 accelerometerBmi160;
    private GyroBmi160 gyroBmi160;
    private Led led;
    private Logging logging;
    private Settings settings;
    private BtleService.LocalBinder serviceBinder;
    private MainActivity mainActivity;

    private DataWriter accelerometerWriter;
    private DataWriter gyroscopeWriter;
    private DataWriter stepWriter;

    static boolean lastTaskIsFinished = true; // Flag to make sure there is no simultaneity allowed.

    private MetawearServices() {
    }

    static MetawearServices getInstance() {

        //if there is no instance available, create new one
        if (metawearServices == null) {
            metawearServices = new MetawearServices();
        }

        return metawearServices;
    }

    /**
     * setup required adapter and serviceBinder.
     *
     * @param bluetoothAdapter bluetooth adapter created in main activity
     * @param serviceBinder    ...
     */
    void setupMetawearServices(MainActivity mainActivity, BluetoothAdapter bluetoothAdapter, BtleService.LocalBinder serviceBinder) {
        this.mainActivity = mainActivity;
        this.bluetoothAdapter = bluetoothAdapter;
        this.serviceBinder = serviceBinder;
        Log.d(TAG, "setupMetawearServices: complete");
    }

    /**
     * start a service which are
     * read information of metawears,
     * start all sensors of all the metawears,
     * stop all sensors of all the metawears as well as remove serialized metawear objects and
     * download data from all the metawears as well as remove all the serialized objects.
     * @param index
     * @param operation
     * @param progress
     */
    void retrieveBoard(int index, Operation operation, ProgressBar progress) {
        String mac = DeviceList.getInstance().getMetawearList().get(index);
        final BluetoothDevice remoteDevice = bluetoothAdapter.getRemoteDevice(mac);

        // Create a MetaWear board object for the Bluetooth Device
        board = serviceBinder.getMetaWearBoard(remoteDevice);

        board.connectAsync().continueWith(task -> {
            if (task.isFaulted()) {
                Log.i("MainActivity", "Failed to connect...reconnecting...");
                retrieveBoard(index, operation, progress);
            } else {
                Log.i("MainActivity", "Connected");

                accelerometerBmi160 = board.getModule(AccelerometerBmi160.class);
                gyroBmi160 = board.getModule(GyroBmi160.class);
                led = board.getModule(Led.class);
                logging = board.getModule(Logging.class);
                settings = board.getModule(Settings.class);

                settings.editBleConnParams()
                        .maxConnectionInterval(Build.VERSION.SDK_INT >= Build.VERSION_CODES.M ? 11.25f : 7.5f)
                        .commit();

                switch (operation) {
                    case SCAN_DEVICE:
                        readBatteryInformation(index);
                        break;
                    case START_DEVICE:
                        startAllSensors();
                        break;
                    case STOP_DEVICE:
                        stopAllSensors(true, true);
                        break;
                    case SYNC_DEVICE:
                        downloadData(progress);
                        break;
                }
            }
            return null;
        });
    }

    // This signal is to show if all processes are finished, if finished, show refresh button.
    private void disconnectBoard(Boolean signal) {
        board.disconnectAsync().continueWith(task -> {
            Log.i("MainActivity", "Disconnected");
            lastTaskIsFinished = true;
            if (signal) {
                mainActivity.runOnUiThread(() -> {
                    mainActivity.refreshButton.show();
                    mainActivity.loading.setVisibility(View.INVISIBLE);
                });
            }
            return null;
        });
    }

    private void readBatteryInformation(int index) {
        board.readBatteryLevelAsync()
                .continueWith(task -> {
                    boolean signal = false;
                    if (task.isFaulted())
                        Log.i(TAG, "Battery Read Faulted on " + board.getMacAddress());
                    else {
                        int batteryLevel = task.getResult().intValue();
                        Log.i("MainActivity", "Battery level for: " + board.getMacAddress() + " ----> " + "%" + batteryLevel);
                        mainActivity.runOnUiThread(() -> {
                            mainActivity.updateBatteryLevel(index, batteryLevel);
                        });
                    }
                    if (index == DeviceList.getInstance().getMetawearList().size() - 1) {
                        signal = true;
                    }
                    disconnectBoard(signal);
                    return null;
                });
    }

    private void startAllSensors() {
        configureAccelerometer(accelerometerBmi160);
        configureGyroscope(gyroBmi160);
        configureStepCounter(accelerometerBmi160);
        accelerometerBmi160.acceleration().addRouteAsync(source -> source.log(ACCELEROMETER_HANDLER)).continueWith((Continuation<Route, Void>) task -> {
            if (task.isFaulted()) {
                Log.i(TAG, "Configuration failed");
                disconnectBoard(true);
            } else
                gyroBmi160.angularVelocity().addRouteAsync(source -> source.log(GYROSCOPE_HANDLER)).continueWith((Continuation<Route, Void>) task1 -> {
                    if (task1.isFaulted()) {
                        Log.i(TAG, "Configuration failed");
                        disconnectBoard(true);
                    } else
                        accelerometerBmi160.stepDetector().addRouteAsync(source -> source.log(STEP_COUNTING_HANDLER)).continueWith((Continuation<Route, Void>) task2 -> {
                            if (task1.isFaulted()) {
                                Log.i(TAG, "Configuration failed");
                                disconnectBoard(true);
                            } else {
                                Log.i(TAG, "Start all sensors");
                                accelerometerBmi160.acceleration().start();
                                accelerometerBmi160.stepDetector().start();
                                accelerometerBmi160.start();

                                gyroBmi160.angularVelocity().start();
                                gyroBmi160.start();

                                serializeBoard();// save current board state

                                //Start logging
                                logging.start(true);
                                Log.i(TAG, "Data collection started for " + board.getMacAddress());
                                led.editPattern(Led.Color.BLUE, Led.PatternPreset.SOLID).commit();
                                led.play();
                                disconnectBoard(true);
                            }
                            return null;
                        });
                    return null;
                });
            return null;
        });
    }

    private void downloadData(ProgressBar progress) throws FileNotFoundException {

        String datetime = DateFormat.getDateTimeInstance().format(new Date());
        accelerometerWriter = new DataWriter(board.getMacAddress() + "_" + datetime + "_accelerometer" + ".csv");
        gyroscopeWriter = new DataWriter(board.getMacAddress() + "_" + datetime + "_gyroscope" + ".csv");
        stepWriter = new DataWriter(board.getMacAddress() + "_" + datetime + "_step" + ".csv");
        stopAllSensors(false, false);
        logging.downloadAsync(100, (nEntriesLeft, totalEntries) -> {
            Log.i("MainActivity", "Progress Update = " + (totalEntries - nEntriesLeft) + "/" + totalEntries);
            mainActivity.runOnUiThread(() -> {
                progress.setMax((int) totalEntries);
                progress.setProgress((int) totalEntries - (int) nEntriesLeft);
            });
        }).continueWithTask((Continuation<Void, Task<Void>>) task -> {
            Log.i("MainActivity", "Download completed");
            progress.setProgress(0);
            board.tearDown();
            disconnectBoard(true);
            accelerometerWriter.closeStream();
            gyroscopeWriter.closeStream();
            stepWriter.closeStream();
            return null;
        });
    }

    private void stopAllSensors(boolean teardown, boolean disconnect) {
        Log.i(TAG, "Stop all sensors");
        accelerometerBmi160.acceleration().stop();
        accelerometerBmi160.stepDetector().stop();
        accelerometerBmi160.stop();

        gyroBmi160.angularVelocity().stop();
        gyroBmi160.stop();
        logging.stop();
        led.stop(true);
        if (teardown)
            board.tearDown();

        deserializeBoard(); // Load current board state for data collection or clear all the serializeFiles

        if (disconnect) {
            disconnectBoard(true);
            return;
        }
        logging = board.getModule(Logging.class);
    }


    /**
     * Static handlers for each subscriber to log to a single file set to env[0]
     */
    private static Subscriber ACCELEROMETER_HANDLER = (Data data, Object... env) -> {
        String value = data.timestamp().getTimeInMillis() + "," +
                data.formattedTimestamp() + "," +
                "0" + "," +
                data.value(Acceleration.class).x() + "," +
                data.value(Acceleration.class).y() + "," +
                data.value(Acceleration.class).z() + "\n";
        getInstance().accelerometerWriter.write(value);
    };

    private static Subscriber GYROSCOPE_HANDLER = (Data data, Object... env) -> {
        String value = data.timestamp().getTimeInMillis() + "," +
                data.formattedTimestamp() + "," +
                "0" + "," +
                data.value(AngularVelocity.class).x() + "," +
                data.value(AngularVelocity.class).y() + "," +
                data.value(AngularVelocity.class).z() + "\n";
        getInstance().gyroscopeWriter.write(value);
    };

    private static Subscriber STEP_COUNTING_HANDLER = (Data data, Object... env) -> {
        String value = data.timestamp().getTimeInMillis() + "\n";
        getInstance().stepWriter.write(value);
    };

    private void configureAccelerometer(AccelerometerBmi160 accelerometerBmi160) {
        accelerometerBmi160.configure()
                .odr(AccelerometerBmi160.OutputDataRate.ODR_50_HZ)       // Set sampling frequency to 50Hz, or closest valid ODR
                .range(AccelerometerBosch.AccRange.AR_4G)
                .commit();
        Log.i(TAG, "Accelerometer Configured");
    }

    /**
     * Sets the Gyroscope Range (2000 degrees) and data rate (50hz)
     */
    private void configureGyroscope(GyroBmi160 gyroBmi160) {
        gyroBmi160.configure()
                .odr(GyroBmi160.OutputDataRate.ODR_50_HZ)
                .range(GyroBmi160.Range.FSR_2000)
                .commit();

        Log.i(TAG, "Gyroscope Configured");
    }

    private void configureStepCounter(AccelerometerBmi160 accelerometerBmi160) {
        accelerometerBmi160.stepDetector()
                .configure()
                .mode(AccelerometerBmi160.StepDetectorMode.NORMAL)
                .commit();
        Log.i(TAG, "Steep Counter Configured");
    }

    /**
     * Serializes the current board into a file locally (identified using the mac address of the device)
     * to be restored when getting data from the device
     */
    public void serializeBoard() {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy_MM_dd");
        String currentDate= sdf.format(new Date());
        File fileDir = new File(Environment.getExternalStorageDirectory().getAbsolutePath(), "mobilityData_"+currentDate);
        if (!fileDir.exists()) {
            if (!fileDir.mkdirs())// create root directory
                Log.i(TAG, "serializeBoard: Failed to create data folder...");
        }
        try {
            File serializeFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath()
                    + File.separator
                    + "mobilityData_"
                    + currentDate
                    + File.separator
                    + board.getMacAddress());
            if (!serializeFile.createNewFile())
                Log.i(TAG, "serializeBoard: Failed to create serializeFile...");
            OutputStream writer = new FileOutputStream(serializeFile, false);
            board.serialize(writer);
            writer.close();
            Log.i(TAG, "Serialized: " + board.getMacAddress());
        } catch (IOException e) {
            Log.i(TAG, "Write failed for: " + board.getMacAddress() + " " + e.toString());
            e.printStackTrace();
        }
    }

    /**
     * Deserializes a board that is already stored on the device
     */
    public void deserializeBoard() {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy_MM_dd");
        String currentDate= sdf.format(new Date());
        try {
            File serializeFile =
                    new File(Environment.getExternalStorageDirectory().getAbsolutePath()
                            + File.separator
                            + "mobilityData_"
                            + currentDate
                            + File.separator
                            + board.getMacAddress());
            if (serializeFile.exists()) {
                InputStream reader = new FileInputStream(serializeFile);
                board.deserialize(reader);
                reader.close();
                if (!serializeFile.delete())
                    Log.i(TAG, "deserializeBoard: Failed to delete serializedFile...");
                Log.i(TAG, "Deserialized: " + board.getMacAddress());
            }
            else {
                mainActivity.runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(mainActivity, "Cannot find serializedBoardFile for: " + board.getMacAddress(), Toast.LENGTH_LONG).show();
                    }
                });
            }
        } catch (IOException | ClassNotFoundException e) {
            Log.i(TAG, "Failed to read: " + board.getMacAddress() + " " + e.toString());
        }
    }

}
